# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from acos_client import client
import acos_client.errors as acos_errors
import responses


HOSTNAME = 'fake_a10'
BASE_URL = "https://{}:443/services/rest/v2.1/?format=json&method=".format(HOSTNAME)
AUTH_URL = "{}authenticate".format(BASE_URL)
CREATE_URL = '{}slb.hm.create&session_id={}'.format(BASE_URL, 'foobar')
DELETE_URL = '{}slb.hm.delete&session_id={}'.format(BASE_URL, 'foobar')
SEARCH_URL = '{}slb.hm.search&session_id={}'.format(BASE_URL, 'foobar')
UPDATE_URL = '{}slb.hm.update&session_id={}'.format(BASE_URL, 'foobar')


class TestHealthMonitor(unittest.TestCase):

    def setUp(self):
        self.client = client.Client(HOSTNAME, '21', 'fake_username', 'fake_password')

    @responses.activate
    def test_health_monitor_create(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            'response': {'status': 'OK'}
        }
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)

        resp = self.client.slb.hm.create('test1', 'HTTP', 5, 5, 5, 'GET', '/', '200', 80)

        self.assertIsNone(resp)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)

    @responses.activate
    def test_health_monitor_create_already_exists(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {
                "code": 2941, "msg": "The same health monitor name already exist."}}
        }
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)

        with self.assertRaises(acos_errors.Exists):
            self.client.slb.hm.create('test1', 'HTTP', 5, 5, 5, 'GET', '/', '200', 80)

        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)

    @responses.activate
    def test_health_monitor_delete(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            'response': {'status': 'OK'}
        }
        responses.add(responses.POST, DELETE_URL, json=json_response, status=200)

        resp = self.client.slb.hm.delete('test1')

        self.assertIsNone(resp)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, DELETE_URL)

    @responses.activate
    def test_health_monitor_delete_not_found(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {"code": 33619968, "msg": " The monitor does not exist."}}
        }
        responses.add(responses.POST, DELETE_URL, json=json_response, status=200)

        resp = self.client.slb.hm.delete('test1')

        self.assertIsNone(resp)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, DELETE_URL)

    @responses.activate
    def test_health_monitor_search(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "health_monitor": {"name": "hfoobar", "retry": 5, "consec_pass_reqd": 5, "interval": 5, "timeout": 5,
                               "strictly_retry": 0, "disable_after_down": 0, "override_ipv4": "0.0.0.0",
                               "override_ipv6": "::", "override_port": 0, "type": 3, "http":
                               {"port": 80, "host": "", "url": "GET /", "user": "", "password": "",
                                "expect_code": "200", "maintenance_code": ""}}}
        responses.add(responses.POST, SEARCH_URL, json=json_response, status=200)

        resp = self.client.slb.hm.get('test1')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, SEARCH_URL)

    @responses.activate
    def test_health_monitor_search_not_found(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {"code": 33619968, "msg": " The monitor does not exist."}}
        }
        responses.add(responses.POST, SEARCH_URL, json=json_response, status=200)

        with self.assertRaises(acos_errors.NotFound):
            self.client.slb.hm.get('test1')

        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, SEARCH_URL)

    @responses.activate
    def test_health_monitor_update(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            'response': {'status': 'OK'}
        }
        responses.add(responses.POST, UPDATE_URL, json=json_response, status=200)

        resp = self.client.slb.hm.update('test1', 'HTTP', 5, 5, 5, 'GET', '/', '200', 80)

        self.assertIsNone(resp)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, UPDATE_URL)

    @responses.activate
    def test_health_monitor_update_not_found(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {"code": 33619968, "msg": " The monitor does not exist."}}
        }
        responses.add(responses.POST, UPDATE_URL, json=json_response, status=200)

        with self.assertRaises(acos_errors.NotFound):
            self.client.slb.hm.update('test1', 'HTTP', 5, 5, 5, 'GET', '/', '200', 80)

        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, UPDATE_URL)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

import json
import logging
from requests.adapters import HTTPAdapter
from requests import Session
import six

import acos_client
from acos_client import logutils
from acos_client.v30 import responses as acos_responses

LOG = logging.getLogger(__name__)

broken_replies = {
    "": '{"response": {"status": "OK"}}'
}


class HttpClient(object):
    HEADERS = {
        "Content-type": "application/json",
        "User-Agent": "ACOS-Client-AGENT-%s" % acos_client.VERSION,
    }

    def __init__(self, host, port=None, protocol="https", max_retries=3, timeout=5):
        if port is None:
            if protocol is 'http':
                self.port = 80
            else:
                self.port = 443
        else:
            self.port = port

        self.url_base = "%s://%s:%s" % (protocol, host, self.port)
        self.max_retries = max_retries
        self.timeout = timeout

    def request(self, method, api_url, params={}, headers=None,
                file_name=None, file_content=None, axapi_args=None, **kwargs):
        LOG.debug("axapi_http: full url = %s", self.url_base + api_url)
        LOG.debug("axapi_http: %s url = %s", method, api_url)
        LOG.debug("axapi_http: params = %s", json.dumps(logutils.clean(params), indent=4))

        valid_http_codes = [200, 204]

        # Update params with axapi_args for currently unsupported configuration of objects
        if axapi_args is not None:
            formatted_axapi_args = dict(
                [(k.replace('_', '-'), v) for k, v in six.iteritems(axapi_args)]
            )
            params = acos_client.v21.axapi_http.merge_dicts(params, formatted_axapi_args)

        # Set data" variable for the request
        if params:
            params_copy = params.copy()
            LOG.debug("axapi_http: params_all = %s", logutils.clean(params_copy))
            payload = json.dumps(params_copy)
        else:
            payload = None

        if (file_name is None and file_content is not None) or \
           (file_name is not None and file_content is None):
            raise ValueError("file_name and file_content must both be populated if one is")

        max_retries = kwargs.get('max_retries', self.max_retries)
        timeout = kwargs.get('timeout', self.timeout)

        # Set "headers" variable for the request
        request_headers = self.HEADERS.copy()
        if headers:
            request_headers.update(headers)
        LOG.debug("axapi_http: headers = %s", json.dumps(logutils.clean(request_headers), indent=4))

        # Process files if passed as a parameter
        if file_name is not None:
            files = {
                'file': (file_name, file_content, "application/octet-stream"),
                'json': ('blob', payload, "application/json")
            }
            request_headers.pop("Content-type", None)
            request_headers.pop("Content-Type", None)

        # Create session to set HTTPAdapter or SSLAdapter and set max_retries
        session = Session()
        if self.port == 443:
            session.mount('https://', HTTPAdapter(max_retries=max_retries))
        else:
            session.mount('http://', HTTPAdapter(max_retries=max_retries))
        session_request = getattr(session, method.lower())

        # Make actual request and handle any errors
        try:
            if file_name is not None:
                device_response = session_request(
                    self.url_base + api_url, verify=False, files=files, headers=request_headers, timeout=timeout
                )
            else:
                device_response = session_request(
                    self.url_base + api_url, verify=False, data=payload, headers=request_headers, timeout=timeout
                )
        except (Exception) as e:
            LOG.error("acos_client failing with error %s after %s retries", e.__class__.__name__, max_retries)
            raise e
        finally:
            session.close()

        # Validate json response
        try:
            json_response = device_response.json()
            LOG.debug("axapi_http: data = %s", json.dumps(logutils.clean(json_response), indent=4))
        except ValueError as e:
            # The response is not JSON but it still succeeded.
            if device_response.status_code in valid_http_codes:
                return device_response.text
            else:
                raise e

        # Handle "fail" responses returned by AXAPI
        if 'response' in json_response and 'status' in json_response['response']:
            if json_response['response']['status'] == 'fail':
                    acos_responses.raise_axapi_ex(json_response, method, api_url)

        # Handle "authorizationschema" responses returned by AXAPI
        if 'authorizationschema' in json_response:
            acos_responses.raise_axapi_auth_error(json_response, method, api_url, headers)

        return json_response

    def get(self, api_url, params={}, headers=None, **kwargs):
        return self.request("GET", api_url, params, headers, **kwargs)

    def post(self, api_url, params={}, headers=None, **kwargs):
        return self.request("POST", api_url, params, headers, **kwargs)

    def put(self, api_url, params={}, headers=None, **kwargs):
        return self.request("PUT", api_url, params, headers, **kwargs)

    def delete(self, api_url, params={}, headers=None, **kwargs):
        return self.request("DELETE", api_url, params, headers, **kwargs)



# Copyright 2014,  Jeff Buttars,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from acos_client import errors as ae
from acos_client.v30 import base


class Action(base.BaseV30):

    def write_memory(self, partition="all", destination="primary", specified_partition=None, **kwargs):
        payload = {
            "memory": {
                "destination": destination,
                "partition": partition
            }
        }

        if specified_partition:
            del payload["memory"]["destination"]
            payload["memory"]["specified-partition"] = specified_partition

        try:
            try:
                self._post("/write/memory/", payload, **kwargs)
            except ae.AxapiJsonFormatError:
                # Workaround regression in 4.1.0 backwards compat
                self._post("/write/memory/", "", **kwargs)
        except ae.ConfigManagerNotReady:
            # If the retry loop missed this, catch it next time.
            pass

    def activate_and_write(self, partition="all", destination="primary", **kwargs):
        self.write_memory(partition, destination)

    def clideploy(self, commandlist, **kwargs):
        payload = {
            "commandlist": commandlist
        }
        return self._post("/clideploy/", payload, **kwargs)

    def reload(self):
        self._post("/reload", "")

    def setInterface(self, interface):
        data = {"ethernet": {"ifnum": str(interface), "name": "DataPort",
                "action": "enable", "ip": {"dhcp": 1}}}
        url = "/interface/ethernet/" + str(interface)
        self._post(url, data)

    def reboot(self):
        self._post("/reboot", "")

    def configureVRRP(self, device_id, set_id):
        data = {"common": {"device-id": device_id, "set-id": set_id,
                "action": "enable"}}
        url = "/vrrp-a/common"
        self._post(url, data)

    def configureVRID(self, vrid):
        data = {"vrid": {"vrid-val": vrid,
                "blade-parameters": {"priority": 150}}}
        url = "/vrrp-a/vrid"
        self._post(url, data)

    def configSynch(self, ip_address, username, password):
        data = {"sync": {"address": ip_address, "auto-authentication": 1,
                "type": "all", "usr": username, "pwd": password}}
        url = "/configure/sync"
        self._post(url, data)

    def set_vcs_device(self, device_id, priority):
        data = {"device": {"device": device_id, "priority": priority,
                "management": 1, "enable": 1}}
        url = "/vcs/device/"
        self._post(url, data)

    def set_vcs_para(self, floating_ip, floating_ip_mask):
        data = {"vcs-para": {"floating-ip-cfg": [{"floating-ip": floating_ip,
                "floating-ip-mask": floating_ip_mask}]}}
        url = "/vcs/vcs-para"
        self._post(url, data)

    def vcs_enable(self):
        data = {"action": {"action": "enable"}}
        url = "/vcs/action"
        self._post(url, data)

    def vcs_reload(self):
        url = "/vcs/reload"
        self._post(url)

    def check_vrrp_status(self):
        url = "/vrrp-a"
        data = self._get(url)
        if "common" in data["vrrp-a"].keys() and \
            "action" in data["vrrp-a"]["common"].keys() and \
                data["vrrp-a"]["common"]["action"] == "enable":
            return True
        else:
            return False

    def get_vcs_summary_oper(self):
        url = "/vcs/vcs-summary/oper"
        return self._get(url)



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

try:
    import unittest
    from unittest import mock
except ImportError:
    import mock
    import unittest2 as unittest

from acos_client.v30.vrrpa import vrid


class TestVRID(unittest.TestCase):
    def setUp(self):
        self.client = mock.MagicMock()
        self.target = vrid.VRID(self.client)
        self.url_prefix = "/axapi/v3/vrrp-a/vrid/"

    def expected_payload(self, vrid_val, threshold=1, disable=0, is_partition=None):
        rv = {
            'vrid': {
                'vrid-val': vrid_val,
                'preempt-mode': {
                    'threshold': threshold,
                    'disable': disable
                }
            }
        }
        return rv

    def test_vrid_get(self):
        self.target.get(0)
        self.client.http.request.assert_called_with("GET", self.url_prefix + '0', {}, mock.ANY)

    def test_vrid_create_threshold(self):
        self.target.create(4, threshold=2)
        self.client.http.request.assert_called_with(
            "POST", self.url_prefix, self.expected_payload(4, threshold=2), mock.ANY)

    def test_vrid_create_disable(self):
        self.target.create(4, disable=1)
        self.client.http.request.assert_called_with(
            "POST", self.url_prefix, self.expected_payload(4, disable=1), mock.ANY)

    def test_vrid_create_floating_ip(self):
        self.target.create(4, threshold=1, disable=0, floating_ips=['10.10.10.8'])
        payload = self.expected_payload(4)
        payload['vrid']['floating-ip'] = mock.ANY
        self.client.http.request.assert_called_with(
            "POST", self.url_prefix, payload,
            mock.ANY)

    def test_vrid_update_threshold(self):
        self.target.update(4, threshold=2)
        self.client.http.request.assert_called_with(
            "PUT", self.url_prefix + '4', self.expected_payload(4, threshold=2), mock.ANY)

    def test_vrid_update_disable(self):
        self.target.update(4, disable=1)
        self.client.http.request.assert_called_with(
            "PUT", self.url_prefix + '4', self.expected_payload(4, disable=1), mock.ANY)

    def test_vrid_update_floating_ip(self):
        self.target.update(4, threshold=1, disable=0, floating_ips=['10.10.10.9'])
        payload = self.expected_payload(4)
        payload['vrid']['floating-ip'] = mock.ANY
        self.client.http.request.assert_called_with(
            "PUT", self.url_prefix + '4', payload,
            mock.ANY)

    def test_patition_vrid_create_floating_ip(self):
        self.target.create(4, threshold=1, disable=0, floating_ips=['10.10.10.8'], is_partition=True)
        payload = self.expected_payload(4, is_partition=True)
        payload['vrid']['floating-ip'] = mock.ANY
        self.client.http.request.assert_called_with(
            "POST", self.url_prefix, payload, mock.ANY)

    def test_partition_vrid_update_floating_ip(self):
        self.target.update(4, threshold=1, disable=0, floating_ips=['10.10.10.9'], is_partition=True)
        payload = self.expected_payload(4, is_partition=True)
        payload['vrid']['floating-ip'] = mock.ANY
        self.client.http.request.assert_called_with(
            "PUT", self.url_prefix + '4', payload, mock.ANY)

    def test_build_params_multi_ip(self):
        floating_ips = ['11.11.11.11', '12.12.12.12', '13.13.13.13']
        floating_ip_payload = [{'ip-address': '11.11.11.11'},
                               {'ip-address': '12.12.12.12'},
                               {'ip-address': '13.13.13.13'}]
        payload = self.target._build_params(0, floating_ips=floating_ips)
        ip_cfg = list(payload['vrid']['floating-ip']['ip-address-cfg'])
        self.assertEqual(floating_ip_payload, ip_cfg)

    def test_build_params_multi_ip_partition(self):
        floating_ips = ['11.11.11.11', '12.12.12.12', '13.13.13.13']
        floating_ip_payload = [{'ip-address-partition': '11.11.11.11'},
                               {'ip-address-partition': '12.12.12.12'},
                               {'ip-address-partition': '13.13.13.13'}]
        payload = self.target._build_params(0, floating_ips=floating_ips, is_partition=True)
        ip_cfg = list(payload['vrid']['floating-ip']['ip-address-part-cfg'])
        self.assertEqual(floating_ip_payload, ip_cfg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from acos_client.v30 import base


class Port(base.BaseV30):

    url_base_tmpl = "/slb/server/{server}/port/"
    url_port_tmpl = "{port}+{protocol}/"

    def create(self, server_name, port, protocol, **kwargs):
        return self._set(server_name, port, protocol, **kwargs)

    def update(self, server_name, port, protocol, **kwargs):
        return self._set(server_name, port, protocol, update=True, **kwargs)

    def delete(self, server_name, port, protocol, **kwargs):
        url = (self.url_base_tmpl + self.url_port_tmpl).format(server=server_name,
                                                               port=port,
                                                               protocol=protocol)

        return self._delete(url)

    def _set(self, server_name, port, protocol, update=False, **kwargs):
        url = self.url_base_tmpl.format(server=server_name)

        params = {
            "port": {
                "conn-resume": kwargs.get("conn_resume", None),
                "conn-limit": kwargs.get("conn_limit"),
                "stats-data-action": kwargs.get("stats_data_action", "stats-data-enable"),
                "weight": kwargs.get("weight", 1),
                "port-number": port,
                "range": kwargs.get("range", 0),
                "action": kwargs.get("action", "enable"),
                "protocol": protocol
            }
        }

        if update:
            url += self.url_port_tmpl.format(port=port, protocol=protocol)

            return self._put(url, params, **kwargs)
        else:
            return self._post(url, params, **kwargs)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import re

from acos_client import errors as ae

RESPONSE_CODES = {
    33619969: {
        '*': {
            '*': ae.InUse
        }
    },
    67371011: {
        '*': {
            '*': ae.Exists
        }
    },
    419495936: {
        '*': {
            '/axapi/v3/logoff': None,
            '*': ae.InvalidSessionID
        }
    },
    520749062: {
        '*': {
            '*': ae.NotFound
        }
    },
    654311495: {
        '*': {
            '*': ae.Exists
        }
    },
    654311505: {
        '*': {
            '*': ae.DhcpAcquireFailed,
        }
    },
    67240011: {
        '*': {
            '*': ae.Exists
        }
    },
    754974732: {
        '*': {
            '*': ae.Exists
        }
    },
    754974733: {
        '*': {
            '*': ae.PartitionIdExists
        }
    },
    1023410176: {
        'DELETE': {
            '*': None
        },
        '*': {
            '*': ae.NotFound
        }
    },
    1023410181: {
        'DELETE': {
            '*': None
        },
        '*': {
            '/axapi/v3/slb/service-group/.*/member/': ae.NotFound,
            '*': ae.NotFound
        }
    },
    1023410183: {
        '*': {
            '*': ae.Exists
        }
    },
    1023451145: {
        '*': {
            '*': ae.Exists
        }
    },
    1023459340: {
        '*': {
            '*': ae.Exists
        }
    },
    1023459393: {
        '*': {
            '*': ae.InvalidParameter
        }
    },
    1023459335: {
        '*': {
            '*': ae.FeatureNotSupported
        }
    },
    1023460352: {
        'DELETE': {
            '*': None
        },
        '*': {
            '*': ae.NotFound
        }
    },
    1023463424: {
        '*': {
            '*': ae.ConfigManagerNotReady
        }
    },
    1023475722: {
        '*': {
            '*': ae.NotFound
        }
    },
    1023508480: {
        '*': {
            '*': ae.AxapiJsonFormatError
        }
    },
    1023509504: {
        '*': {
            '*': ae.NotFound
        }
    },
    1023524874: {
        '*': {
            '*': ae.AxapiJsonFormatError
        }
    },
    1023656960: {
        '*': {
            '*': ae.NotFound
        }
    },
    1023656962: {
        '*': {
            '*': ae.NotFound
        }
    },
    1207960052: {
        '*': {
            '/axapi/v3/logoff': None,
            '*': ae.InvalidSessionID
        }
    },
    1207959957: {
        '*': {
            '*': ae.NotFound
        }
    },
    1208025092: {
        '*': {
            '/axapi/v3/logoff': None,
            '*': ae.InvalidSessionID
        }
    },
    1208025095: {
        '*': {
            '*': ae.ConfigManagerNotReady
        }
    },
    1208078344: {
        '*': {
            '*': ae.NotFound
        }
    },
    1023443968: {
        'DELETE': {
            '*': None
        },
        '*': {
            '*': ae.NotFound
        }
    },
    1023451144: {
        '*': {
            '*': ae.Exists
        }
    },
    1023459337: {
        '*': {
            '*': ae.Exists
        }
    },
    1023459339: {
        '*': {
            '*': ae.Exists
        }
    },
    1023475727: {
        '*': {
            '*': ae.NotFound
        }
    },
    1208008960: {
        "*": {
            "*": ae.AuthenticationFailure
        }
    },
    4294967295: {
        '*': {
            '*': ae.ConfigManagerNotReady
        }
    },
}


def raise_axapi_auth_error(response, method, api_url, headers):
    if 'authorizationschema' in response:
        code = response['authorizationschema']['code']
        s = response['authorizationschema']['error']
        if code == 401:
            if headers and 'Authorization' in headers:
                raise ae.InvalidSessionID(code, s)
            else:
                raise ae.AuthenticationFailure(code, s)
        elif code == 403:
            raise ae.AuthenticationFailure(code, s)


def raise_axapi_ex(response, method, api_url):
    if 'response' in response and 'err' in response['response']:
        code = response['response']['err']['code']

        # Check if this is a known error code that we want to map.
        if code in RESPONSE_CODES:
            ex_dict = RESPONSE_CODES[code]
            ex = None

            # Now match against specific HTTP method exceptions
            if method in ex_dict:
                x = ex_dict[method]
            else:
                x = ex_dict['*']

            # Now try to find specific API method exceptions
            matched = False
            for k in x.keys():
                if k != '*' and re.match('^' + k, api_url):
                    matched = True
                    ex = x[k]

            # If we get here, try for a fallback exception for this code
            if not matched and not ex and '*' in x:
                ex = x['*']

            # Alright, time to actually do something
            if ex:
                raise ex(code, response['response']['err']['msg'])
            else:
                return

        raise ae.ACOSException(code, response['response']['err']['msg'])

    raise ae.ACOSException()



# Copyright 2014-2016, A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals
import six

from acos_client.v30 import base
from acos_client.v30.slb.virtual_port import VirtualPort


class VirtualServer(base.BaseV30):
    url_prefix = '/slb/virtual-server/'

    @property
    def vport(self):
        return VirtualPort(self.client)

    def all(self):
        return self._get(self.url_prefix)

    def get(self, name):
        return self._get(self.url_prefix + name)

    def _set(self, name, ip_address=None, arp_disable=False, description=None, vrid=None,
             virtual_server_templates=None, template_virtual_server=None,
             port_list=None, **kwargs):
        params = {
            "virtual-server": self.minimal_dict({
                "name": name,
                "ip-address": ip_address,
                "arp-disable": None if arp_disable is None else int(arp_disable),
                "description": description,
                "port-list": port_list
            }),
        }
        if self._is_ipv6(ip_address):
            params['virtual-server']['ipv6-address'] = ip_address
        else:
            params['virtual-server']['ip-address'] = ip_address

        if description:
            params['virtual-server']['description'] = description
        else:
            params['virtual-server']['description'] = None

        if vrid:
            params['virtual-server']['vrid'] = int(vrid)
        if virtual_server_templates:
            virtual_server_templates = {k: v for k, v in virtual_server_templates.items() if v}
            params['virtual-server']['template-virtual-server'] = \
                virtual_server_templates.get('template-virtual-server', None)
            params['virtual-server']['template-logging'] = virtual_server_templates.get('template-logging', None)
            params['virtual-server']['template-policy'] = virtual_server_templates.get('template-policy', None)
            params['virtual-server']['template-scaleout'] = virtual_server_templates.get('template-scaleout', None)

        # for backward compatibility
        if template_virtual_server:
            params['virtual-server']['template-virtual-server'] = str(template_virtual_server)

        config_defaults = kwargs.get("config_defaults")
        if config_defaults:
            for k, v in six.iteritems(config_defaults):
                params['virtual-server'][k] = v

        return params

    def create(self, name, ip_address, arp_disable=False, description=None, vrid=None,
               virtual_server_templates=None, template_virtual_server=None,
               port_list=None, **kwargs):
        params = self._set(name, ip_address, arp_disable=arp_disable, description=description,
                           vrid=vrid, virtual_server_templates=virtual_server_templates,
                           template_virtual_server=template_virtual_server,
                           port_list=port_list, **kwargs)
        return self._post(self.url_prefix, params, **kwargs)

    def update(self, name, ip_address=None, arp_disable=False, description=None, vrid=None,
               virtual_server_templates=None, template_virtual_server=None,
               port_list=None, **kwargs):
        params = self._set(name, ip_address, arp_disable=arp_disable, description=description,
                           vrid=vrid, virtual_server_templates=virtual_server_templates,
                           template_virtual_server=template_virtual_server,
                           port_list=port_list, **kwargs)
        return self._post(self.url_prefix + name, params, **kwargs)

    def replace(self, name, ip_address=None, arp_disable=False, description=None, vrid=None,
                virtual_server_templates=None, template_virtual_server=None,
                port_list=None, **kwargs):
        params = self._set(name, ip_address, arp_disable=arp_disable, description=description,
                           vrid=vrid, virtual_server_templates=virtual_server_templates,
                           template_virtual_server=template_virtual_server,
                           port_list=port_list, **kwargs)
        return self._put(self.url_prefix + name, params, **kwargs)

    def delete(self, name):
        return self._delete(self.url_prefix + name)

    def stats(self, name='', **kwargs):
        resp = self._get(self.url_prefix + name + '/port/stats', **kwargs)
        return resp

    def oper(self, name='', **kwargs):
        resp = self._get(self.url_prefix + name + '/oper', **kwargs)
        return resp



# Copyright 2014-2016, A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import json
import responses

try:
    import unittest
except ImportError:
    import unittest2 as unittest

from acos_client import client
import acos_client.errors as acos_errors


HOSTNAME = 'fake_a10'
BASE_URL = 'https://{}:443/axapi/v3'.format(HOSTNAME)
AUTH_URL = '{}/auth'.format(BASE_URL)
VSERVER_NAME = 'test'
CREATE_URL = '{}/slb/virtual-server/'.format(BASE_URL)
OBJECT_URL = '{}/slb/virtual-server/{}'.format(BASE_URL, VSERVER_NAME)
STATS_URL = '{}/slb/virtual-server/{}/port/stats'.format(BASE_URL, VSERVER_NAME)
OPER_URL = '{}/slb/virtual-server/{}/oper'.format(BASE_URL, VSERVER_NAME)


class TestVirtualServer(unittest.TestCase):

    def setUp(self):
        self.client = client.Client(HOSTNAME, '30', 'fake_username', 'fake_password')

    @responses.activate
    def test_virtual_server_create_no_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 0,
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.create('test', '192.168.2.254')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_create_with_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 1,
                'description': 'test_description',
                'vrid': 1,
                'template-virtual-server': 'TEST_VIP_TEMPLATE',
            }
        }

        resp = self.client.slb.virtual_server.create(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            description='test_description',
            vrid=1,
            template_virtual_server='TEST_VIP_TEMPLATE',
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_create_with_existing_template(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 1,
                'vrid': 1,
                'template-virtual-server': 'template_sv',
                'template-logging': 'template_lg',
                'template-policy': 'template_pl',
                'template-scaleout': 'template_sc',
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.create(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            vrid=1,
            virtual_server_templates={
                'template-virtual-server': 'template_sv',
                'template-logging': 'template_lg',
                'template-policy': 'template_pl',
                'template-scaleout': 'template_sc',
            }
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_create_with_non_existing_template(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 1,
                'vrid': 1,
                'template-virtual-server': 'template_sv2',
                'template-logging': None,
                'template-policy': None,
                'template-scaleout': None,
            }
        }

        resp = self.client.slb.virtual_server.create(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            vrid=1,
            virtual_server_templates={
                'template-virtual-server': 'template_sv',
            }
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertNotEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_update_no_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 0,
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.update('test', '192.168.2.254')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_update_with_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'name': VSERVER_NAME,
                'ip-address': '192.168.2.254',
                'arp-disable': 1,
                'description': 'test_description',
                'vrid': 1,
                'template-virtual-server': 'TEST_VIP_TEMPLATE',
            }
        }

        resp = self.client.slb.virtual_server.update(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            description='test_description',
            vrid=1,
            template_virtual_server='TEST_VIP_TEMPLATE'
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_update_with_existing_template(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'name': VSERVER_NAME,
                'ip-address': '192.168.2.254',
                'arp-disable': 1,
                'vrid': 1,
                'template-virtual-server': 'TEST_VIP_TEMPLATE',
                'template-logging': 'template_lg',
                'template-policy': None,
                'template-scaleout': None,
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.update(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            vrid=1,
            virtual_server_templates={'template-logging': 'template_lg'}, template_virtual_server='TEST_VIP_TEMPLATE',
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_replace_no_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.PUT, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
                'arp-disable': 0,
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.replace('test', '192.168.2.254')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.PUT)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_replace_with_params(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.PUT, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'name': VSERVER_NAME,
                'ip-address': '192.168.2.254',
                'arp-disable': 1,
                'description': 'test_description',
                'vrid': 1,
                'template-virtual-server': 'TEST_VIP_TEMPLATE',
            }
        }

        resp = self.client.slb.virtual_server.replace(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            description='test_description',
            vrid=1,
            template_virtual_server='TEST_VIP_TEMPLATE'
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.PUT)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_replace_with_existing_template(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.PUT, OBJECT_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'name': VSERVER_NAME,
                'ip-address': '192.168.2.254',
                'arp-disable': 1,
                'vrid': 1,
                'template-virtual-server': 'TEST_VIP_TEMPLATE',
                'template-logging': 'template_lg',
                'template-policy': None,
                'template-scaleout': None,
                'description': None,
            }
        }

        resp = self.client.slb.virtual_server.replace(
            name='test',
            ip_address='192.168.2.254',
            arp_disable=1,
            vrid=1,
            virtual_server_templates={'template-logging': 'template_lg'}, template_virtual_server='TEST_VIP_TEMPLATE',
        )

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.PUT)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_virtual_server_delete(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.DELETE, OBJECT_URL, json=json_response, status=200)

        resp = self.client.slb.virtual_server.delete('test')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.DELETE)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)

    @responses.activate
    def test_virtual_server_delete_not_found(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {"code": 67239937, "msg": " No such Virtual Server"}}
        }
        responses.add(responses.DELETE, OBJECT_URL, json=json_response, status=200)

        with self.assertRaises(acos_errors.ACOSException):
            self.client.slb.virtual_server.delete('test')

        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.DELETE)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)

    @responses.activate
    def test_virtual_server_search(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.GET, OBJECT_URL, json=json_response, status=200)

        resp = self.client.slb.virtual_server.get('test')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)

    @responses.activate
    def test_virtual_server_search_not_found(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            "response": {"status": "fail", "err": {"code": 67239937, "msg": " No such Virtual Server"}}
        }
        responses.add(responses.GET, OBJECT_URL, json=json_response, status=200)

        with self.assertRaises(acos_errors.ACOSException):
            self.client.slb.virtual_server.get('test')

        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, OBJECT_URL)

    @responses.activate
    def test_virtual_server_stats(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.GET, STATS_URL, json=json_response, status=200)

        resp = self.client.slb.virtual_server.stats('test')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, STATS_URL)

    @responses.activate
    def test_virtual_server_oper(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.GET, OPER_URL, json=json_response, status=200)

        resp = self.client.slb.virtual_server.oper('test')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, OPER_URL)

    @responses.activate
    def test_virtual_server_create_no_arp_disable(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {"foo": "bar"}
        responses.add(responses.POST, CREATE_URL, json=json_response, status=200)
        params = {
            'virtual-server': {
                'ip-address': '192.168.2.254',
                'name': VSERVER_NAME,
            }
        }

        resp = self.client.slb.virtual_server.create('test', '192.168.2.254')

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertTrue("arp-disable" not in params["virtual-server"])



# Copyright 2014,  Jeff Buttars,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Session(object):

    def __init__(self, client, username, password):
        self.client = client
        self.http = client.http
        self.username = username
        self.password = password
        self.session_id = None

    @property
    def id(self):
        if self.session_id is None:
            self.authenticate(self.username, self.password)
        return self.session_id

    def authenticate(self, username, password):
        url = "/axapi/v3/auth"
        payload = {
            'credentials': {
                "username": username,
                "password": password
            }
        }

        if self.session_id is not None:
            self.close()

        r = self.http.post(url, payload)
        if "authresponse" in r:
            self.session_id = str(r['authresponse']['signature'])
        else:
            self.session_id = None

        return r

    def close(self):
        try:
            self.client.partition.active()
        except Exception:
            pass

        if self.session_id is None:
            return

        try:
            h = {'Authorization': "A10 %s" % self.session_id}
            r = self.http.post('/axapi/v3/logoff', headers=h)
        finally:
            self.session_id = None

        return r



# Copyright 2016, All Rights Reserved,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import six

try:
    import unittest
    from unittest import mock
except ImportError:
    import mock
    import unittest2 as unittest

from acos_client import logutils as target


class TestLogutils(unittest.TestCase):
    def setUp(self):
        self.clean_fields = ["username", "password"]

        self.obj_flat = self._fakeobj(a=1, b=2, username="admin", password="secret")
        self.obj_onelevel = self._fakeobj(a=1,
                                          b=2,
                                          credentials=self._fakeobj(
                                              username="admin",
                                              password="secret"))
        self.obj_twolevel = self._fakeobj(a=1,
                                          b=2,
                                          credentials=self._fakeobj(
                                              inside_secret=self._fakeobj(
                                                  username="admin",
                                                  password="secret")))
        self.flat_dict = {
            "a": 1,
            "b": 2,
            "username": "admin",
            "password": "secret"
        }

        self.dict_onelevel = {
            "a": 1,
            "b": 2,
            "credentials": {
                "username": "admin",
                "password": "secret"
            }
        }

        self.dict_twolevel = {
            "a": 1,
            "b": 2,
            "credentials": {
                "inside_secret": {
                    "username": "admin",
                    "password": "secret"
                }
            }
        }

    def _fakeobj(self, *args, **kwargs):
        return FakeObject(**kwargs)

    def test_clean_flat_dict(self):
        actual = target.clean(self.flat_dict)
        self.assertEqual(target.REPLACEMENT, actual["username"])
        self.assertEqual(target.REPLACEMENT, actual["password"])
        self.assertEqual(self.flat_dict["a"], actual["a"])

    def test_clean_onelevel_dict(self):
        actual = target.clean(self.dict_onelevel).get("credentials")

        self.assertEqual(target.REPLACEMENT, actual["username"])
        self.assertEqual(target.REPLACEMENT, actual["password"])
        self.assertEqual(1, self.dict_onelevel["a"])

    def test_clean_twolevel_dict(self):
        actual = target.clean(self.dict_twolevel).get("credentials").get("inside_secret")

        self.assertEqual(target.REPLACEMENT, actual["username"])
        self.assertEqual(target.REPLACEMENT, actual["password"])

    def test_tuple_dict(self):
        actual = target.clean(
            (
                1,
                {
                    'credentials': {
                        'username': 'admin',
                        'password': 'secret'
                    }
                }
            )
        )
        expected = (
            1,
            {
                'credentials': {
                    'username': target.REPLACEMENT,
                    'password': target.REPLACEMENT
                }
            }
        )
        self.assertEqual(expected, actual)

    def test_list_dict(self):
        actual = target.clean(
            [
                {
                    'credentials': {
                        'username': 'admin',
                        'password': 'secret'
                    }
                }
            ]
        )
        expected = [
            {
                'credentials': {
                    'username': target.REPLACEMENT,
                    'password': target.REPLACEMENT
                }
            }
        ]
        self.assertEqual(expected, actual)

    def test_int(self):
        actual = target.clean(1)
        self.assertEqual(1, actual)

    def test_float(self):
        actual = target.clean(3.7)
        self.assertEqual(3.7, actual)

    def test_none(self):
        actual = target.clean(None)
        self.assertEqual(None, actual)

    def test_string(self):
        actual = target.clean('sometext')
        self.assertEqual('sometext', actual)

    def test_ustring(self):
        actual = target.clean('sometext')
        self.assertEqual('sometext', actual)

    def test_mock(self):
        # It's likely that clean will be called with a mock during testing.
        # It shouldn't blow up

        m = mock.MagicMock()
        str(m)
        target.clean(m)


class FakeObject(object):
    def __init__(self, *args, **kwargs):
        for k, v in six.iteritems(kwargs):
            setattr(self, k, v)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals
import json
import responses

try:
    import unittest
except ImportError:
    import unittest2 as unittest

from acos_client import client

HOSTNAME = 'fake_a10'
BASE_URL = 'https://{}:443/axapi/v3'.format(HOSTNAME)
AUTH_URL = '{}/auth'.format(BASE_URL)
MEM_NAME = 'fake-server'
SG_NAME = 'fake-sg'
CREATE_URL = '{}/slb/service-group/{}/member/'.format(BASE_URL, SG_NAME)
OBJECT_URL = '{}/slb/service-group/{}/member/{}+{}/'
OK_RESP = {'response': {'status': 'OK'}}


class TestMember(unittest.TestCase):

    def setUp(self):
        self.client = client.Client(HOSTNAME, '30', 'fake_username', 'fake_password')
        self.member = self.client.slb.service_group.member

    @responses.activate
    def test_create_enable_member(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        responses.add(responses.POST, CREATE_URL, json=OK_RESP, status=200)
        params = {
            'member': {
                'name': MEM_NAME,
                'port': 80,
                'member-stats-data-disable': self.member.STATUS_ENABLE,
                'member-state': 'enable',
            }
        }
        resp = self.member.create(SG_NAME, MEM_NAME, 80)

        self.assertEqual(resp, OK_RESP)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_create_disable_member(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        responses.add(responses.POST, CREATE_URL, json=OK_RESP, status=200)
        params = {
            'member': {
                'name': MEM_NAME,
                'port': 80,
                'member-stats-data-disable': self.member.STATUS_DISABLE,
                'member-state': 'disable',
            }
        }
        resp = self.member.create(SG_NAME, MEM_NAME, 80,
                                  self.member.STATUS_DISABLE,
                                  False)

        self.assertEqual(resp, OK_RESP)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, CREATE_URL)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_update_enable_member(self):
        port = 443
        object_url = OBJECT_URL.format(BASE_URL, SG_NAME, MEM_NAME, port)
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        responses.add(responses.POST, object_url, json=OK_RESP, status=200)
        params = {
            'member': {
                'name': MEM_NAME,
                'port': port,
                'member-stats-data-disable': self.member.STATUS_ENABLE,
                'member-state': 'enable',
            }
        }
        resp = self.member.update(SG_NAME, MEM_NAME, port)

        self.assertEqual(resp, OK_RESP)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, object_url)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_update_disable_member(self):
        port = 443
        object_url = OBJECT_URL.format(BASE_URL, SG_NAME, MEM_NAME, port)
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        responses.add(responses.POST, object_url, json=OK_RESP, status=200)
        params = {
            'member': {
                'name': MEM_NAME,
                'port': port,
                'member-stats-data-disable': self.member.STATUS_DISABLE,
                'member-state': 'disable',
            }
        }
        resp = self.member.update(SG_NAME, MEM_NAME, port,
                                  self.member.STATUS_DISABLE,
                                  False)

        self.assertEqual(resp, OK_RESP)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.POST)
        self.assertEqual(responses.calls[1].request.url, object_url)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)

    @responses.activate
    def test_replace_enable_member(self):
        port = 443
        object_url = OBJECT_URL.format(BASE_URL, SG_NAME, MEM_NAME, port)
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        responses.add(responses.PUT, object_url, json=OK_RESP, status=200)
        params = {
            'member': {
                'name': MEM_NAME,
                'port': port,
                'member-stats-data-disable': self.member.STATUS_ENABLE,
                'member-state': 'enable',
            }
        }
        resp = self.member.replace(SG_NAME, MEM_NAME, port)

        self.assertEqual(resp, OK_RESP)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.PUT)
        self.assertEqual(responses.calls[1].request.url, object_url)
        self.assertEqual(json.loads(responses.calls[1].request.body), params)



# Copyright 2015,  Tobit Raff,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals
import six


from acos_client import errors as acos_errors
from acos_client.v30 import base


class BaseL7(base.BaseV30):

    def get(self, name, **kwargs):
        return self._get(self.url_prefix + name, **kwargs)

    def exists(self, name):
        try:
            self.get(name)
            return True
        except acos_errors.NotFound:
            return False

    def _set(self, name, insert_client_ip=0, insert_client_ip_header_name="", client_ip_hdr_replace=0,
             request_header_insert_list=[], response_header_insert_list=[], update=False,
             **kwargs):
        # Unimplemented options:
        # Everything except for insert_client_ip and request header insert.

        obj_params = {
            "name": name,
            "insert-client-ip": insert_client_ip,
            "insert-client-ip-header-name": insert_client_ip_header_name,
            "client-ip-hdr-replace": client_ip_hdr_replace,
            "request-header-insert-list": request_header_insert_list,
            "response-header-insert-list": response_header_insert_list
        }

        params = {'%s' % self.prefix: {}}
        for key, val in six.iteritems(obj_params):
            # Filter out invalid, or unset keys
            if val != "":
                params['%s' % self.prefix][key] = val
        if not update:
            name = ''
        self._post(self.url_prefix + name, params, **kwargs)

    def create(self, name, insert_client_ip=False, insert_client_ip_header_name="",
               client_ip_hdr_replace=0, request_header_insert_list=[], response_header_insert_list=[],
               **kwargs):
        if self.exists(name):
            raise acos_errors.Exists

        self._set(name, insert_client_ip, insert_client_ip_header_name, client_ip_hdr_replace,
                  request_header_insert_list, response_header_insert_list, **kwargs)

    def update(self, name, insert_client_ip=False, insert_client_ip_header_name="",
               client_ip_hdr_replace=0, request_header_insert_list=[], response_header_insert_list=[],
               **kwargs):
        self._set(name, insert_client_ip, insert_client_ip_header_name, client_ip_hdr_replace,
                  request_header_insert_list, response_header_insert_list, update=True, **kwargs)

    def delete(self, name, **kwargs):
        self._delete(self.url_prefix + name, **kwargs)


class HTTPTemplate(BaseL7):

    url_prefix = '/slb/template/http/'
    prefix = 'http'



# Copyright 2014,  Jeff Buttars,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import ipaddress
import six
import time

from acos_client import errors as ae


class BaseV30(object):

    def __init__(self, client):
        self.client = client
        self.http = client.http
        self.auth_header = {}

    def minimal_dict(self, my_dict, exclude=[]):
        return dict((k, v) for k, v in my_dict.items() if v is not None or k in exclude)

    def url(self, action):
        self.auth_header['Authorization'] = "A10 %s" % self.client.session.id
        return ("/axapi/v3" + action)

    def _request(self, method, action, params, retry_count=0, **kwargs):
        if retry_count > 24:
            raise ae.ACOSUnknownError()

        try:
            return self.client.http.request(method, self.url(action), params,
                                            self.auth_header, **kwargs)
        except (ae.InvalidSessionID, ae.ConfigManagerNotReady) as e:
            if type(e) == ae.ConfigManagerNotReady:
                retry_limit = 24
                sleep_secs = 5
            else:
                retry_limit = 5
                sleep_secs = 1.0

            if retry_count < retry_limit:
                time.sleep(sleep_secs)
                try:
                    p = self.client.current_partition
                    self.client.session.close()
                    self.client.partition.active(p)
                except Exception:
                    pass
                return self._request(method, action, params, retry_count + 1, **kwargs)
            raise e

    def _get(self, action, params={}, **kwargs):
        return self._request('GET', action, params, **kwargs)

    def _post(self, action, params={}, **kwargs):
        return self._request('POST', action, params, **kwargs)

    def _put(self, action, params={}, **kwargs):
        return self._request('PUT', action, params, **kwargs)

    def _delete(self, action, params={}, **kwargs):
        return self._request('DELETE', action, params, **kwargs)

    def _is_ipv6(self, ip_address):
        validated_ip_address = ipaddress.ip_address(six.text_type(ip_address))
        return isinstance(validated_ip_address, ipaddress.IPv6Address)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from acos_client import client
import responses


HOSTNAME = 'fake_a10'
BASE_URL = "https://{}:443/services/rest/v2.1/?format=json&method=".format(HOSTNAME)
AUTH_URL = "{}authenticate".format(BASE_URL)
SYS_INFO_URL = '{}system.information.get&session_id={}'.format(BASE_URL, 'foobar')
SYS_WRITE_MEM_URL = '{}system.action.write_memory&session_id={}'.format(BASE_URL, 'foobar')


class TestHighAvailability(unittest.TestCase):

    def setUp(self):
        self.client = client.Client(HOSTNAME, '21', 'fake_username', 'fake_password')

    @responses.activate
    def test_system_information(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {
            'system_information': {
                'advanced_core_os_on_compact_flash1': 'No Software',
                'advanced_core_os_on_compact_flash2': 'No Software',
                'advanced_core_os_on_harddisk1': '2.7.1-P3-AWS(build: 4)',
                'advanced_core_os_on_harddisk2': '2.7.1-P3-AWS(build: 4)',
                'aflex_engine_version': '2.0.0',
                'axapi_version': '2.1',
                'current_time': '03:25:47 IST Tue Jul 1 2014',
                'firmware_version': 'N/A',
                'last_config_saved': '06:25:26 GMT Sat Dec 28 2013',
                'serial_number': 'N/A',
                'software_version': '2.7.1-P3-AWS(build: 4)',
                'startup_mode': 'hard disk primary',
                'technical_support': 'www.a10networks.com/support '
            }
        }
        responses.add(responses.GET, SYS_INFO_URL, json=json_response, status=200)

        resp = self.client.system.information()

        self.assertEqual(resp, json_response)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, SYS_INFO_URL)

    @responses.activate
    def test_system_write_memory(self):
        responses.add(responses.POST, AUTH_URL, json={'session_id': 'foobar'})
        json_response = {'response': {'status': 'OK'}}
        responses.add(responses.GET, SYS_WRITE_MEM_URL, json=json_response, status=200)

        resp = self.client.system.action.write_memory()

        self.assertIsNone(resp)
        self.assertEqual(len(responses.calls), 2)
        self.assertEqual(responses.calls[1].request.method, responses.GET)
        self.assertEqual(responses.calls[1].request.url, SYS_WRITE_MEM_URL)



# Copyright 2016, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from acos_client.v21 import base


class Interface(base.BaseV21):
    def _build_payload(self, ifnum=None, ip_address=None, ip_netmask=None, dhcp=False, enable=None,
                       speed="auto", **kwargs):
        cli_cmds = []

        if ifnum:
            cli_cmds += ["interface ethernet {port_num}".format(port_num=ifnum)]
        if ip_address and not dhcp:
            cli_cmds += ['ip address {ip} {mask}'.format(ip=ip_address,
                                                         mask=ip_netmask)]
        elif dhcp:
            cli_cmds += ['ip address dhcp']
        cli_cmds += ['write mem']
        return "\n".join(cli_cmds)

    def get_list(self):
        return self._get('network.interface.getAll')

    def get(self, ifnum=None):
        params = {
            "port_num": ifnum
        }
        return self._post('network.interface.get', params=params)

    def delete(self, ifnum):
        raise NotImplementedError("Not implemented in AXAPI v2.1")

    def create(self, ifnum, ip_address=None, ip_netmask=None, dhcp=False, enable=None,
               speed="auto"):
        payload = self._build_payload(ifnum=ifnum, ip_address=ip_address, ip_netmask=ip_netmask,
                                      dhcp=dhcp, enabled=enable, speed=speed)
        return self._post("cli.deploy", params=None,
                          payload=payload)

    def update(self, ifnum, ip_address=None, ip_netmask=None, dhcp=False, enable=None,
               speed="auto"):
        payload = self._build_payload(ifnum=ifnum, ip_address=ip_address, ip_netmask=ip_netmask,
                                      dhcp=dhcp, enable=enable, speed=speed)
        return self._request("POST", "cli.deploy", params=None, payload=payload)

    @property
    def ethernet(self):
        return EthernetInterface(self.client)

    @property
    def management(self):
        return ManagementInterface(self.client)


class EthernetInterface(Interface):
    pass


class ManagementInterface(Interface):
    def _build_payload(self, ifnum=None, ip_address=None, ip_netmask=None, ip_gateway=None,
                       dhcp=False, enable=None, speed="auto", apps_use_mgmt_port=True, **kwargs):
        # TODO(mdurrant) - Check ip/netmask for validity.
        rv = {
            "status": 1,
        }

        rv["speed"] = speed
        rv["apps_use_mgmt_port"] = 1 if apps_use_mgmt_port else 0

        if ip_address and not dhcp:
            rv["ipv4_address"] = ip_address
            rv["ipv4_netmask"] = ip_netmask
            rv["ipv4_gateway"] = ip_gateway
        else:
            rv["ipv4_address"] = "dhcp"

        return {"mgmt_interface": rv}

    def get(self):
        return self._get('network.mgmt_interface.get')

    def create(self, ifnum=None, ip_address=None, ip_netmask=None, dhcp=False, enable=None,
               speed="auto", default_gateway=None):
        payload = self._build_payload(ifnum=ifnum, ip_address=ip_address, ip_netmask=ip_netmask,
                                      dhcp=dhcp, enabled=enable, speed=speed,
                                      default_gateway=default_gateway)
        return self._post("network.mgmt_interface.set",
                          payload)

    def update(self, ifnum=None, ip_address=None, ip_netmask=None, dhcp=False, enable=None,
               speed="auto", default_gateway=None):
        payload = self._build_payload(ifnum=ifnum, ip_address=ip_address, ip_netmask=ip_netmask,
                                      dhcp=dhcp, enable=enable, speed=speed,
                                      default_gateway=default_gateway)
        return self._post("network.mgmt_interface.set",
                          payload)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from acos_client.v21 import base


class VRRPAInterface(base.BaseV21):
    def get_all(self, **kwargs):
        return self._get("vrrp_a.interface.getAll", **kwargs)

    def search(self, interface_num):
        params = {"interface_number": interface_num}

        return self._post("vrrp_a.interface.search", params)

    def update(self, interface_number, interface_type, status, vrrp_a_status, link_type, heartbeat, vlan=None):
        env_val = "vrrp_a_interface"
        params = {
            env_val: {
                "interface_number": interface_number,
                "interface_type": interface_type,
                "status": status,
                "vrrp_a_status": vrrp_a_status,
                "link_type": link_type,
                "heartbeat": heartbeat
            }
        }

        if vlan:
            params[env_val]["vlan"] = vlan

        return self._post("vrrp_a.interface.update", params)
